;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "emacsql" "4.4.1"
  "High-level SQL database front-end."
  '((emacs "28.1"))
  :url "https://github.com/magit/emacsql"
  :commit "d811bbefcb5e27841af55cae53aa939ba720de77"
  :revdesc "v4.4.1-0-gd811bbefcb5e"
  :authors '(("Christopher Wellons" . "wellons@nullprogram.com"))
  :maintainers '(("Jonas Bernoulli" . "emacs.emacsql@jonas.bernoulli.dev")))
